function [est_src]=CD_Ext_Infomax(xold,in_struct)
%=================================================================================
%This function implements the unconstrained and constrained extended
%Infomax algorithm as described in the General Non-orthogonal constrained ICA
%paper (http://ieeexplore.ieee.org/xpl/articleDetails.jsp?arnumber=6802439)
%For a better description of the input parameters, please read the paper.
%This function uses the absolute value of the correlation coefficient as
%the closeness metric instead of the dot product.
%In other words, the closeness threshold chosen by the user
%should be between 0 and 1./
%
%I. Inputs:
%1. xold = Mixture data (it does not have to be squared)
%2. in_struct = structure with various parameters that are needed,
%including the following ones:
% in_struct.no_white = if "1" do not perform whitening in here, any other value
% performs whitening (this is an obligatory input)
% in_struct.tolerance = provide tolerance to determine algorithm convergence (default 1e-4)
% in_struct.constraint_flag  = if equal to "5" apply constrained version of the
% algorithm, if  "0" run un-constrained version.
% in_struct.mu = initial step parameter for constrained algorithm update (defaults to 0.5)
% in_struct.guess_mat = matrix where each column is a reference vector for one of the columns in A (X=AS)
% in_struct.err1 = correlation threshold for constrained algorithm (default 0.7)
% in_struct.lm = lagrange multiplier step parameter (defaults to 3)
% in_struct.inv_Q = Inverse of the whitening matrix used to whiten xold,
% (it assumes in_struct.no_white =0 and in_struct.constraint_flag =5).
% if data is not whitened outside then it will happen inside here (see in_struct.no_white =1);
%
%II. Outputs:
%est_src = estimated sources
%
%III. How To use (giveng mixture data: X)
%  whitesig = Q * X;%where X=mixture data, and Q is calculated whitening matrix
%  inv_Q=pinv(Q);%inverse whitening matrix
%
% in_struct.no_white=1;
% in_struct.tolerance=1e-3;
% in_struct.constraint_flag=5;
% in_struct.mu=0.5;
% in_struct.guess_mat=guess_mat; %
% in_struct.err1=0.9;
% in_struct.lm=3;
% in_struct.inv_Q=inv_Q;
%
% [est_src]=CD_Ext_Infomax(whitesig,in_struct);
%
%If you just want to run un-constrained version, run this:
% in_struct.no_white=1;
% in_struct.tolerance=1e-3;
% in_struct.constraint_flag=0; 
%
% [est_src]=CD_Ext_Infomax(whitesig,in_struct);
%
% To test code call function with no input arguments: CD_Ext_Infomax
% For this demo we simulate an ICA problem, by using ten natural images as sources and
% randomly creating a 10x10 mixing matrix. We use one of the mixing vectors
% as a constraint. We run the un-constained and the constrained version of the Extended
% Infomax algorithm. For more details of the algorithm, see description
% insed CD_Ext_Infomax.m.
%
% Coded by Pedro A. Rodriguez (rodripe1 at umbc.edu)
% Final version and modifications by Zois Boukouvalas (zb1 at umbc.edu)
%
% Refernces:
% [1] Rodriguez, P.A.; Anderson, M.; Xi-Lin Li; Adali, T., "General
% Non-Orthogonal Constrained ICA," IEEE Transactions on Signal Processing, vol.62, no.11, pp.2778-2786, June 2014.

%% Call internal test function when no arguments supplied
if nargin==0
   help CD_Ext_Infomax
   test_CD_Ext_Infomax
   return
end
%%

%Start of original function:----------------------------------------------------
[n,m] = size(xold);

%Reading some of the parameters provided by user or seeting the defaults:
max_iter_alg=1000;
if isfield(in_struct,'max_iter_alg')
    max_iter_alg=in_struct.max_iter_alg;
end
defl = 1;

tolerance = 1e-4;
if isfield(in_struct,'tolerance');
    tolerance=in_struct.tolerance;
end

constraint_flag =0; %if "1" then do natural gradient (default == 0)
if isfield(in_struct,'constraint_flag')
    constraint_flag=in_struct.constraint_flag;
end

%To peform whitening, or not that is the question:
no_white=in_struct.no_white;
if no_white == 1
    X=xold;
    Q=diag(ones(1,n));
elseif no_white == 2
    yyy = zeros(1,m);
    [Ex, Dx] = eig(cov(xold'));
    Ex=diag(ones(size(Ex,1),1));
    Q = mtimes(sqrt(inv(Dx)),Ex');
    X = Q * xold;
else
    % Whitening of s:
    yyy = zeros(1,m);
    [Ex, Dx] = eig(cov(xold'));
    Q = mtimes(sqrt(inv(Dx)),Ex');
    X = Q * xold;
end
cov_X=cov(X');
Q_1=pinv(Q);


%Weight matrix:
if 1 || constraint_flag <= 1
    W=eye(n);
else
    W = randn(n,n);
    W = symdecor(W);
end

if constraint_flag == 5 || constraint_flag == 6  
    QQ=Q;
    if isfield(in_struct,'inv_Q');
        Q_1=in_struct.inv_Q*Q_1;
        QQ=pinv(Q_1);
        
    end
    
    guess=in_struct.guess_mat;
    num_guess=size(guess,2);
    
    num_W=size(W,1);
    %Re-sort existing W based on correlation with matrix W:
    
    for kl=1:num_guess
        r_i=guess(:,kl);
        for lp=1:num_W
            w=W(lp,:).';
            R_par = corrcoef(Q_1*w,r_i);
            corr_w_guess(kl,lp)=R_par(1,2);
        end
    end
    
    %We may need to use auction to chose order:
    [maxi,max_idx]=max(abs(corr_w_guess),[],2); % ; Added by Zois 
    
    if length(unique(max_idx)) ~= num_guess
        [colsol, rowsol] = auction((1-abs(corr_w_guess))');
        max_idx=colsol';
    end
    
    c = setxor((1:num_W).', max_idx);
    sort_order=[max_idx ;c];
    
    W=W(sort_order,:);
    
end


del_W=W.*0;
last_W=W;

y_i=randn(n,1);

mu=0.5;
if isfield(in_struct,'mu');
    mu=in_struct.mu;
end

mu_y=0.5;
if isfield(in_struct,'mu_y');
    mu_y=in_struct.mu_y;
end


%Loop through different iterations:
for iter = 1 : max_iter_alg
    
    
    %Calculate initial estimate of sources:
    Y = W*X;
    G=zeros(size(Y));
    dG=G;
    
    
    for ij=1:n
        
        % Decoupling trick:
        % if n is too large, we adopt a smarter way to calculate h_n to
        % reduce the computational load
        if n>7
            
            if ij==1
                Wn = W(2:n,:);
                inv_Q = inv( Wn*Wn' );
            else
                n_last = ij-1;
                Wn_last = [ W(1:n_last-1,:);W(n_last+1:n,:) ];
                w_current = W(ij,:)';
                w_last = W(n_last,:)';
                c = Wn_last*( w_last - w_current );
                c(n_last) = 0.5*( w_last'*w_last - w_current'*w_current );
                e_last = zeros(n-1,1);
                e_last(n_last) = 1;
                
                temp1 = inv_Q*c;
                temp2 = inv_Q(:,n_last);
                inv_Q_plus = inv_Q - temp1*temp2'/(1+temp1(n_last));
                
                temp1 = inv_Q_plus'*c;
                temp2 = inv_Q_plus(:,n_last);
                inv_Q = inv_Q_plus - temp2*temp1'/(1+c'*temp2);
                % inv_Q is Hermitian
                inv_Q = (inv_Q+inv_Q')/2;
            end
            
            temp1 = randn(n, 1);
            W_n = [ W(1:ij-1,:);W(ij+1:n,:) ];
            h = temp1 - W_n'*inv_Q*W_n*temp1;
            
        else
            temp1 = randn(n, 1);
            temp2 = [W(1:ij-1,:); W(ij+1:n,:)];
            h = temp1 - temp2'*inv(temp2*temp2')*temp2*temp1;
        end
        h_n=h;
        
        w=W(ij,:)';
        y = w'*X;
        
        %nonlinearity selection:
        scl_Y=y;
        % scl_Y=y./std(y);
        y_i(ij)=(1-mu_y).*y_i(ij) + mu_y*mean((-tanh(scl_Y).*scl_Y + (1-tanh(scl_Y).^2)));
        
        if y_i(ij) > 0 || constraint_flag == 2 || constraint_flag == 3
            G(ij,:)=-2*tanh(y);
            dG(ij,:)=-2*(1-tanh(y).^2); %derivative in terms of "y"
        else
            G(ij,:)=tanh(y)-y;
            dG(ij,:)=(1-tanh(y).^2)-1; %derivative in terms of "y"
        end
        g=G(ij,:);
        d_g=dG(ij,:); %derivative
        
        spc_t=h_n./(h_n'*w); %special term
        d_spc_t=-1.*(h_n*h_n')./(h_n'*w).^2; %derivative
        
        if constraint_flag == 5  %constrained with correlation after un-whitening w
            guess=in_struct.guess_mat;
            num_guess= size(guess, 2);
            
            if (ij <= num_guess)
                
                r_i=guess(:,ij);
                
                
                if iter == 1
                    err1=0.7;
                    if isfield(in_struct,'err1');
                        err1=in_struct.err1;
                    end
                    
                    lm=3;%penalty term
                    if isfield(in_struct,'lm');
                        lm=in_struct.lm;
                    end
                    
                    mu1(ij)=(1-err1)*lm; %lagrange multiplier (inequality constrain,mse criteria)
                    mu1(ij)=0;
                else
                    %
                end
                
                R_par = corrcoef(Q_1*w,r_i); %correlation
                r_corr=abs(R_par(1,2));
                mu1_new(ij)=max(0,mu1(ij) + lm*(err1-r_corr));
                mu1(ij)=sign(R_par(1,2))*mu1_new(ij);
                
                %Modify r_i:
                r_i_mu=QQ*r_i;
                r_i_mu=r_i_mu/norm(r_i_mu);
                
                
                w=w + mu.*(spc_t+ ((g*X')./m)' + mu1(ij).*r_i_mu );
            else
                w=w + mu.*(spc_t+ ((g*X')./m)');
                
            end
            
            
            
        else
            %Update demixing vector
            w=w + mu.*(spc_t+ ((g*X')./m)');%Maybe multiply with B'*B
            
            
        end
        
        
        
        w = w / norm(w);
        
        %Update big W:
        W(ij,:)=w';
        
    end
    
    
    %Decide to stop or not based on tolerance
    oldwtchange = W-last_W;
    delta=reshape(oldwtchange,1,n*n);
    change=delta*delta';
    
    if iter>1 && change< tolerance
        break
    else
        last_W=W;
    end
    
    
end

est_src=W*X; %Estimated source
est_W=(pinv(Q)*W.').'; %Estimated demixing matrix

return;

function test_CD_Ext_Infomax
%% built-in test function for constrained ICA
disp('Executing internal test function for CD_Ext_Infomax.')

%Load natural images:
load('natural_images_128x128_gray.mat');%loading variable "images"
orig_images=images;
normal_images=images(:,:,[1:5 8:12]);% images(:,:,6) and images(:,:,7) are noise and not useful

%Some variables to be setup by user:
num_iter=1; %number of iterations to run test code, only thing that changes is the random mixing matrix
guess_num=[2]; %We are going to use the truth mixing vector of the second source for the constrained algorigthm. This is just an example, perfect ground truth is not available in real life ;)
ds=1; %Scale factor to resize images (if desired)



%% Start running code
num_alg=2;
alg_trick=[1 2];

%% Create Sources: S_zero_mean
[m_normal,n_normal,d_normal]=size(normal_images);

m=round(ds*m_normal);
n=round(ds*n_normal);
d=d_normal;

S=zeros(d,m*n);
data=[];
fonti_size=1;
for ij=1:d
    h1=figure();
    
    img=imresize(normal_images(:,:,ij),[m n],'bilinear');
    
    subplot(1,2,1),imagesc(img),colormap('gray'),axis tight,axis off,
    
    subplot(1,2,2),hist(img(:),128);
    set(gca,'FontSize',fonti_size),xlabel('Pixel Histogram Bins','FontSize',fonti_size),
    ylabel('Number of Pixels','FontSize',fonti_size);
    
    %Save statistical information about the image:
    data(ij).skewness=skewness(img(:));
    data(ij).kurtosis=kurtosis(img(:));
    data(ij).cen_mom_4=moment(img(:),4);
    data(ij).mom_4=dot(img(:).^4,ones(m*n,1))./(m*n);
    data(ij).mean=mean(img(:));
    data(ij).std=std(img(:));
    data(ij).median=median(img(:));
    
    %Make every image zero mean:
    img=img-mean2(img);
    
    %Create source matrix for ICA run.
    S(ij,:)=reshape(img,1,m*n);
    
    
end
close all;



%Remove overall mean of new source matrix:
tempVar = mean(S,2)*ones(1,size(S, 2));
S_zero_mean = S - tempVar;
mean_of_S=tempVar;
clear tempVar;

%%
numsig=d; %number of sources equals number of images
A_coll=zeros(numsig,numsig,num_iter);
for iter=1:num_iter
    
    %Create random mixing matrix for every iteration
    k=0;
    while (k ~= numsig)
        MIX_MATRIX=randn(numsig);
        c = cond(MIX_MATRIX);
        k = rank(MIX_MATRIX);
        A_coll(:,:,iter)=MIX_MATRIX;
    end
    
    %% Create ICA mixture matrix (concatenate rows of images)
    X=MIX_MATRIX* S; %X=AS
    
    %Remove mean of new mixture matrix (this step is not necessary in this example, since the sources were already zero mean)
    tempVar = mean(X,2)*ones(1,size(X, 2));
    X_zero_mean = X - tempVar;
    mean_of_X_trans=tempVar;
    clear tempVar;
    
    %Saving ground truth data at the end S_total structure:
    S_total(num_alg+1).icasig=S;
    S_total(num_alg+1).size=[m,n];
    S_total(num_alg+1).A=MIX_MATRIX;
    S_total(num_alg+1).W=inv(MIX_MATRIX);
    S_total(num_alg+1).max_idx=(1:numsig)';
    S_total(num_alg+1).pro_dec=0;
    
    %Run unconstrained and constrained extended infomax:
    %For meaning of input parameters in "in_struct" see definitions inside
    %CD_Ext_Infomax.m
    
    for alg=1:num_alg %number of iterations 
        in_struct=[];
        switch alg_trick(alg)
            case 1
                in_struct.no_white=0;
                in_struct.tolerance=1e-3;
                in_struct.constraint_flag=0;
                in_struct.mu=0.5;
                
                icasig_info=CD_Ext_Infomax(X_zero_mean,in_struct);
                icasig=icasig_info;
                
                
                
                
            case 2 %Using constrained version
                guess_mat= (S_total(num_alg+1).A(:,guess_num));%OR?
                
                in_struct.no_white=0;
                in_struct.tolerance=1e-3;
                in_struct.constraint_flag=5;
                in_struct.mu=0.5;
                in_struct.guess_mat=guess_mat;
                in_struct.err1=0.9;
                in_struct.lm=3;
                
                icasig_info=CD_Ext_Infomax(X_zero_mean,in_struct);
                icasig=icasig_info;
                
                
                
        end
        
        %Re-shape estimated sources into 2D matrices (display them by un-commenting lines inside for loop):
        S_hat=zeros(d,m*n);
        for ij=1:numsig
            S_hat(ij,:)=icasig(ij,:);
            
            %             S_hat_img=reshape(S_hat(ij,:),m,n);
            %             figure(ij+10)
            %             subplot(1,2,1),imagesc(S_hat_img),colormap('gray'),axis equal
            %             subplot(1,2,2),hist(S_hat(ij,:))
        end
        %         close all;
        
        %Find correspondence between estimated and original sources:
        zzz=cat(1,S,S_hat);
        R = corrcoef(zzz');
        R_imp=abs(R(1:numsig,numsig+1:end));
        
        max_val=zeros(numsig,1);
        max_idx=zeros(numsig,1);
        for ij=1:numsig
            [max_val(ij) max_idx(ij)]=max(R_imp(ij,:));
        end
        
        %In cases were source estimation was not one-to-one, we use auction algorithm
        %for assignment.
        pro_dec=0;
        if (length(unique(max_idx)) < numsig)
            %                 pause
            disp(['Using auction algorithm to determine correspondence between truth and estimated sources in alg ' num2str(alg) ' and iter ' num2str(iter)]);
            pro_dec=1;
            
            [colsol, rowsol] = auction((1-R_imp)');
            max_idx=colsol';
            
            for ij=1:numsig
                max_val(ij)=R_imp(ij,max_idx(ij));
            end
            
        end
        
        
        %Saving recovered data for each algorithm:
        S_total(alg).icasig=icasig;
        S_total(alg).size=[m,n];
        S_total(alg).max_idx=max_idx;
        S_total(alg).max_val=max_val;
        S_total(alg).pro_dec=pro_dec;
        
        
        
        close all
    end
    
    %Compare correlation coefficient of constrained and unconstrained estimated
    %sources with the ground truth source (this humber should be
    %positive most of the time).
    improve_correlation(iter)=mean(S_total(2).max_val(guess_num)-S_total(1).max_val(guess_num));
    mean_improvement=mean(improve_correlation);
    disp(['The average improvment in the correlation coefficient is ' num2str(mean_improvement)])
end
return




%Additional functions (we dont use all of them)
function W=symdecor(M)
%fast symmetric orthogonalization
[V D]=eig(M*M');
W=(V.*(ones(size(M,2),1)*(1./sqrt(diag(D)'))))*V'*M;
return


function [X, P] = pre_processing( X )
% pre-processing program
[N,T] = size(X);
% remove DC
Xmean=mean(X,2);
X = X - Xmean*ones(1,T);

% spatio pre-whitening 1
R = X*X'/T;
P = inv_sqrtmH(R);  %P = inv(sqrtm(R));
X = P*X;
return


function A = inv_sqrtmH( B )
%
[V,D] = eig(B);
d = diag(D);
d = 1./sqrt(d);
A = V*diag(d)*V';
return


function [colsol, rowsol] = auction(assignCost)

%function [colsol, rowsol] = auction(assignCost, guard)
%AUCTION : performs assignment using Bertsekas' auction algorithm
%          The auction is 1-sided and uses a fixed epsilon.
%  colsol = auction(assignCost) returns column assignments: colsol(j) gives the
%           row assigned to column j.
%
%           assignCost is an m x n matrix of costs for associating each row
%           with each column.  m >= n (rectangular assignment is allowed).
%           Auction finds the assignment that minimizes the costs.
%
%  colsol = auction(assignCost, guard) sets the cost of column
%           non-assignment = guard.  All assignments will have cost < guard.
%           A column will not be assigned if the savings from eliminating it
%           exceeds the guard cost.  colsol(j) = 0 indicates that the optimal
%           solution does not assign column j to any row.
%
%  [colsol,rowsol] = auction(assignCost) also returns the row assignments:
%           rowsol(i) gives the column assigned to row j.
%
%  Reference
%  Bertsekas, D. P., "The Auction Algorithm: A Distributed Relaxation Method
%  for the Assignment Problem," Annals of Operations Research, 14, 1988, pp.
%  103-123.

% Mark Levedahl
% 
%=================================================================================

[m,n] = size(assignCost);

if m < n
	error('cost matrix must have no more columns than rows.')
end

% augment cost matrix with a guard row if specified.
m0 = m;
% if isfinite(guard) %%% DONE BY JS
% 	m = m+1;
% 	assignCost(m,:) = guard;
% end

% init return arrays
colsol = zeros(1,n);
rowsol = zeros(m,1);
price = zeros(m,1);
EPS = sqrt(eps) / (n+1);

% 1st step is a full parallel solution.  Get bids for all columns
jp = 1:n;
f = assignCost;
[b1,ip] = min(f);                   % cost and row of best choice
f(ip + m*(0:n-1)) = inf;            % eliminate best from contention
bids = min(f) - b1;                 % cost of runner up choice hence bid

% Arrange bids so highest (best) are last and will overwrite the lower bids.
[tmp,ibid] = sort(bids(:));

% Now assign best bids (lesser bids are overwritten by better ones).
price(ip(ibid)) = price(ip(ibid)) + EPS + tmp;
rowsol(ip(ibid)) = jp(ibid);        % do row assignments
iy = find(rowsol);
colsol(rowsol(iy)) = iy;            % map to column assignments

% The guard row cannot be assigned (always available)
if m0 < m
	price(m) = 0;
	rowsol(m) = 0;
end

% Now Continue with non-parallel code handling any contentions.
while ~all(colsol)
	for jp = find(~colsol)
		f = assignCost(:,jp) + price;   % costs
		[b1,ip] = min(f);               % cost and row of best choice
		if ip > m0
			colsol(jp) = m;
		else
			f(ip) = inf;                    % eliminate from contention
			price(ip) = price(ip) + EPS + min(f) - b1; % runner up choice hence bid
			if rowsol(ip)                   % take the row away if already assigned
				colsol(rowsol(ip)) = 0;
			end
			rowsol(ip) = jp;                % update row and column assignments
			colsol(jp) = ip;
		end % if ip == m
	end
end

% screen out infeasible assignments
if m > m0
	colsol(colsol == m) = 0;
	rowsol(m) = [];
end
return